from __future__ import annotations

import logging
from typing import (
    Any,
    Callable,
    Dict,
    Mapping,
    Optional,
    Type,
    Union,
    overload,
    Literal,
)
from typing import Tuple

from pyspark.sql import DataFrame, SparkSession

from dc43_integrations.spark.io.common import (
    GovernanceSparkWriteRequest,
)
from dc43_integrations.spark.violation_strategy import (
    NoOpWriteViolationStrategy,
    WriteViolationStrategy,
)
from dc43_integrations.spark.io.base import (
    BaseWriteExecutor,
    WriteExecutionResult,
)
from dc43_service_clients.governance.client.interface import GovernanceServiceClient
from dc43_service_clients.governance.models import GovernanceWriteContext, GovernancePolicy

logger = logging.getLogger(__name__)


def write_with_governance(
    df: DataFrame,
    request: GovernanceSparkWriteRequest | Mapping[str, object],
    *,
    governance_service: GovernanceServiceClient,
    enforce: bool = True,
    auto_cast: bool = True,
    streaming_batch_callback: Optional[Callable[[Mapping[str, Any]], None]] = None,
) -> WriteExecutionResult:
    """Write a DataFrame allowing governance service to resolve policy."""

    if not isinstance(request, GovernanceSparkWriteRequest):
        if isinstance(request, Mapping):
            request = GovernanceSparkWriteRequest(**dict(request))
        else:
            raise TypeError("request must be a GovernanceSparkWriteRequest or mapping")

    violation_strategy = request.violation_strategy or NoOpWriteViolationStrategy()
    context = request.context
    if not isinstance(context, GovernanceWriteContext):
        context = GovernanceWriteContext(**dict(context))
    
    if context.policy is None:
        context.policy = GovernancePolicy()

    if (
        context.policy.allowed_data_product_statuses is None
        and isinstance(violation_strategy, WriteViolationStrategy)
    ):
        allowed = violation_strategy.allowed_data_product_statuses
        if isinstance(allowed, str):
            context.policy.allowed_data_product_statuses = (allowed,)
        else:
            context.policy.allowed_data_product_statuses = tuple(allowed)
    if (
        context.policy.allow_missing_data_product_status is None
        and isinstance(violation_strategy, WriteViolationStrategy)
    ):
        context.policy.allow_missing_data_product_status = bool(
            violation_strategy.allow_missing_data_product_status
        )
    if (
        context.policy.data_product_status_case_insensitive is None
        and isinstance(violation_strategy, WriteViolationStrategy)
    ):
        context.policy.data_product_status_case_insensitive = bool(
            violation_strategy.data_product_status_case_insensitive
        )
    if (
        context.policy.data_product_status_failure_message is None
        and isinstance(violation_strategy, WriteViolationStrategy)
    ):
        failure_message = violation_strategy.data_product_status_failure_message
        if failure_message is not None:
            context.policy.data_product_status_failure_message = str(failure_message)
    if context.policy.enforce_data_product_status is None:
        context.policy.enforce_data_product_status = bool(enforce)

    plan = governance_service.resolve_write_context(context=context)
    pipeline_ctx = request.context.pipeline_context or plan.pipeline_context

    executor = BaseWriteExecutor(
        df=df,
        request=request,
        governance_service=governance_service,
        enforce=enforce,
        auto_cast=auto_cast,
        streaming_batch_callback=streaming_batch_callback,
        plan=plan,
    )
    return executor.execute()




"""Local helpers for exercising contract-decorated DLT assets with databricks-dlt."""

from __future__ import annotations

from dataclasses import dataclass
from functools import wraps
from types import ModuleType
from typing import Any, Callable, Dict, Mapping, TypeVar, TYPE_CHECKING

try:  # pragma: no cover - optional dependency guard
    import dlt as databricks_dlt
except Exception as exc:  # pragma: no cover - databricks-dlt not installed
    databricks_dlt = None  # type: ignore[assignment]
    _DLT_IMPORT_ERROR = exc
else:  # pragma: no cover - only executed when the import succeeds
    _DLT_IMPORT_ERROR = None


def _build_stub_dlt_module() -> ModuleType:
    """Return a minimal ``dlt`` facsimile for environments without the package."""

    module = ModuleType("dc43_integrations.stub_dlt")
    module.__dc43_is_stub__ = True  # type: ignore[attr-defined]

    # ------------------------------------------------------------------
    # Runtime toggles
    # ------------------------------------------------------------------
    def enable_local_execution() -> None:  # pragma: no cover - no-op
        return None

    module.enable_local_execution = enable_local_execution  # type: ignore[attr-defined]

    # ------------------------------------------------------------------
    # Expectation decorators
    # ------------------------------------------------------------------
    def _noop_decorator(fn: F) -> F:  # pragma: no cover - trivial
        return fn

    def expect_all(predicates: Mapping[str, str]) -> Callable[[F], F]:
        return _noop_decorator

    def expect_all_or_drop(predicates: Mapping[str, str]) -> Callable[[F], F]:
        return _noop_decorator

    module.expect_all = expect_all  # type: ignore[attr-defined]
    module.expect_all_or_drop = expect_all_or_drop  # type: ignore[attr-defined]

    # ------------------------------------------------------------------
    # Asset decorators
    # ------------------------------------------------------------------
    def table(query_function: F | None = None, **kwargs: Any) -> Callable[[F], F]:
        if query_function is not None and callable(query_function):
            return query_function

        def decorator(fn: F) -> F:
            return fn

        return decorator

    def view(query_function: F | None = None, **kwargs: Any) -> Callable[[F], F]:
        if query_function is not None and callable(query_function):
            return query_function

        def decorator(fn: F) -> F:
            return fn

        return decorator

    module.table = table  # type: ignore[attr-defined]
    module.view = view  # type: ignore[attr-defined]

    return module


_STUB_DLT_MODULE: Any | None = None


def ensure_dlt_module(*, allow_stub: bool = False) -> Any:
    """Return the ``dlt`` module or a stub replacement when requested."""

    if databricks_dlt is not None:
        return databricks_dlt
    if not allow_stub:
        raise RuntimeError(
            "databricks-dlt package is required for DLT helpers"
        ) from _DLT_IMPORT_ERROR

    global _STUB_DLT_MODULE
    if _STUB_DLT_MODULE is None:
        _STUB_DLT_MODULE = _build_stub_dlt_module()
    return _STUB_DLT_MODULE

try:  # pragma: no cover - optional dependency guard
    from pyspark.errors import AnalysisException
    from pyspark.sql.functions import expr
except Exception as exc:  # pragma: no cover - pyspark not installed
    AnalysisException = RuntimeError  # type: ignore[assignment]
    expr = None  # type: ignore[assignment]
    _PYSPARK_IMPORT_ERROR = exc
else:  # pragma: no cover - only executed when pyspark is present
    _PYSPARK_IMPORT_ERROR = None

if TYPE_CHECKING:  # pragma: no cover - typing only
    from pyspark.sql import DataFrame, SparkSession
else:  # pragma: no cover - runtime placeholder
    DataFrame = Any  # type: ignore[assignment]
    SparkSession = Any  # type: ignore[assignment]


F = TypeVar("F", bound=Callable[..., "DataFrame"])


@dataclass(frozen=True)
class ExpectationReport:
    """Single expectation evaluation emitted during local DLT execution."""

    asset: str
    rule: str
    predicate: str
    action: str
    failed_rows: int
    run_id: int

    @property
    def status(self) -> str:
        """Return ``"passed"`` when ``failed_rows`` equals zero."""

        return "passed" if self.failed_rows == 0 else "failed"


class LocalDLTHarness:
    """Patch ``databricks-dlt`` so contract-decorated assets can run locally.

    The upstream `databricks-dlt` package only toggles a ``LOCAL_EXECUTION_MODE``
    flag; the expectation and table decorators still return a stub ``__outer``
    function that yields ``None``.  Tests therefore need a thin shim that
    registers the decorated callables, executes them against a Spark session and
    records the expectation verdicts.  ``LocalDLTHarness`` fills that gap while
    reusing the official decorators so notebooks remain source-compatible with
    Databricks deployments.
    """

    def __init__(self, spark: "SparkSession", *, module: Any | None = None) -> None:
        if expr is None:  # pragma: no cover - exercised when pyspark is missing
            raise RuntimeError(
                "LocalDLTHarness requires pyspark; install pyspark to run local DLT tests"
            ) from _PYSPARK_IMPORT_ERROR
        if module is None:
            module = ensure_dlt_module(allow_stub=True)
        self.spark = spark
        self.module = module
        self._using_stub = bool(getattr(module, "__dc43_is_stub__", False))
        self.expectation_reports: list[ExpectationReport] = []
        self.table_options: Dict[str, Mapping[str, Any]] = {}
        self.view_options: Dict[str, Mapping[str, Any]] = {}
        self._tables: Dict[str, Callable[[], "DataFrame"]] = {}
        self._views: Dict[str, Callable[[], "DataFrame"]] = {}
        self._current_asset: str | None = None
        self._run_sequence = 0
        self._originals: Dict[str, Any] = {}
        self._active = False

    # ------------------------------------------------------------------
    # Lifecycle management
    # ------------------------------------------------------------------
    def activate(self) -> None:
        """Patch the underlying ``dlt`` module so decorators use the harness."""

        if self._active:
            return
        enable = getattr(self.module, "enable_local_execution", None)
        if callable(enable):  # pragma: no branch - trivial guard
            enable()
        self._originals = {
            name: getattr(self.module, name)
            for name in ("expect_all", "expect_all_or_drop", "table", "view")
        }
        setattr(self.module, "expect_all", self._expect_all)
        setattr(self.module, "expect_all_or_drop", self._expect_all_or_drop)
        setattr(self.module, "table", self._table)
        setattr(self.module, "view", self._view)
        self._active = True

    def deactivate(self) -> None:
        """Restore the original ``dlt`` decorators."""

        if not self._active:
            return
        for name, original in self._originals.items():
            setattr(self.module, name, original)
        self._originals.clear()
        self._active = False

    def __enter__(self) -> "LocalDLTHarness":  # pragma: no cover - simple guard
        self.activate()
        return self

    def __exit__(self, exc_type, exc, tb) -> None:  # pragma: no cover - simple guard
        self.deactivate()

    # ------------------------------------------------------------------
    # Expectation decorators
    # ------------------------------------------------------------------
    def _expect_all(self, predicates: Mapping[str, str]) -> Callable[[F], F]:
        return self._build_expectation_decorator(predicates, action="warn", drop=False)

    def _expect_all_or_drop(self, predicates: Mapping[str, str]) -> Callable[[F], F]:
        return self._build_expectation_decorator(predicates, action="drop", drop=True)

    def _build_expectation_decorator(
        self,
        predicates: Mapping[str, str],
        *,
        action: str,
        drop: bool,
    ) -> Callable[[F], F]:
        predicate_map = dict(predicates)

        def decorator(fn: F) -> F:
            @wraps(fn)
            def wrapper(*args: Any, **kwargs: Any) -> "DataFrame":
                df = fn(*args, **kwargs)
                if df is None:
                    raise RuntimeError("DLT asset returned None; expected a DataFrame")
                asset_name = self._resolve_asset_name(wrapper, fn)
                for key, predicate in predicate_map.items():
                    try:
                        invalid = df.filter(~expr(predicate))
                        failures = invalid.count()
                    except AnalysisException:
                        failures = df.count()
                    self._record(asset_name, key, predicate, action, failures)
                return df

            return wrapper  # type: ignore[return-value]

        return decorator

    # ------------------------------------------------------------------
    # Asset registration
    # ------------------------------------------------------------------
    def _table(self, query_function: F | None = None, **kwargs: Any) -> Callable[[F], F]:
        if query_function is not None and callable(query_function):
            return self._register_asset(query_function, kwargs, self._tables, self.table_options)

        def decorator(fn: F) -> F:
            return self._register_asset(fn, kwargs, self._tables, self.table_options)

        return decorator

    def _view(self, query_function: F | None = None, **kwargs: Any) -> Callable[[F], F]:
        if query_function is not None and callable(query_function):
            return self._register_asset(query_function, kwargs, self._views, self.view_options)

        def decorator(fn: F) -> F:
            return self._register_asset(fn, kwargs, self._views, self.view_options)

        return decorator

    def _register_asset(
        self,
        fn: F,
        options: Mapping[str, Any],
        registry: Dict[str, Callable[[], "DataFrame"]],
        option_store: Dict[str, Mapping[str, Any]],
    ) -> F:
        name = str(options.get("name") or getattr(fn, "__name__", "asset"))
        setattr(fn, "__dlt_asset_name__", name)
        registry[name] = fn
        option_store[name] = dict(options)
        return fn

    # ------------------------------------------------------------------
    # Execution helpers
    # ------------------------------------------------------------------
    def run_asset(self, name: str) -> "DataFrame":
        """Execute a registered table or view and return the resulting DataFrame."""

        if name in self._tables:
            fn = self._tables[name]
        elif name in self._views:
            fn = self._views[name]
        else:
            raise KeyError(f"unknown DLT asset: {name}")

        previous_asset = self._current_asset
        self._current_asset = name
        self._run_sequence += 1
        try:
            result = fn()
        finally:
            self._current_asset = previous_asset
        if result is None:
            raise RuntimeError(f"DLT asset {name} returned None; expected a DataFrame")
        return result

    def run_all(self) -> Dict[str, "DataFrame"]:
        """Execute all registered tables and return a mapping of results."""

        return {name: self.run_asset(name) for name in self._tables}

    # ------------------------------------------------------------------
    # Internal helpers
    # ------------------------------------------------------------------
    def _resolve_asset_name(self, wrapper: Callable[..., Any], original: Callable[..., Any]) -> str:
        name = getattr(wrapper, "__dlt_asset_name__", None)
        if not name:
            name = getattr(original, "__dlt_asset_name__", None)
        return str(name or getattr(original, "__name__", "asset"))

    def _record(self, asset: str, rule: str, predicate: str, action: str, failures: int) -> None:
        report = ExpectationReport(
            asset=asset,
            rule=rule,
            predicate=predicate,
            action=action,
            failed_rows=failures,
            run_id=self._run_sequence,
        )
        self.expectation_reports.append(report)


__all__ = ["ExpectationReport", "LocalDLTHarness", "ensure_dlt_module"]



from __future__ import annotations

import logging
import tempfile
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Mapping,
    Optional,
    Protocol,
    Sequence,
)

from pyspark.sql import DataFrame

from dc43_service_clients.data_quality import ValidationResult, ObservationPayload
from dc43_service_clients.governance.client.interface import GovernanceServiceClient
from open_data_contract_standard.model import OpenDataContractStandard

from dc43_integrations.spark.data_quality import collect_observations
from dc43_integrations.spark.io.common import (
    _safe_fs_name,
)
from dc43_service_backends.core.odcs import contract_identity

logger = logging.getLogger(__name__)


def _derive_metrics_checkpoint(
    base: Optional[str],
    dataset_id: Optional[str],
    dataset_version: Optional[str],
) -> str:
    """Return a checkpoint path for streaming metric collectors."""

    if isinstance(base, str) and base:
        trimmed = base.rstrip("/")
        if trimmed.endswith("_dq"):
            return trimmed
        return f"{trimmed}_dq"

    safe_id = _safe_fs_name(dataset_id or "stream")
    safe_version = _safe_fs_name(dataset_version or "unknown")  # Use "unknown" if None to match original logic approx
    # Original used _timestamp() if None:
    # safe_version = _safe_fs_name(dataset_version or _timestamp())
    # I need _timestamp. I can import it from strategies? No, circular.
    # Duplicate strict timestamp logic or utility?
    # Or just use datetime.utcnow() directly here.
    if not dataset_version:
        now = datetime.utcnow().replace(tzinfo=timezone.utc)
        safe_version = _safe_fs_name(now.isoformat().replace("+00:00", "Z"))
    
    root = Path(tempfile.gettempdir()) / "dc43_stream_metrics" / safe_id / safe_version
    try:
        root.mkdir(parents=True, exist_ok=True)
    except Exception:  # pragma: no cover
        pass
    return str(root)


class StreamingInterventionError(RuntimeError):
    """Raised when a streaming intervention strategy blocks the pipeline."""


@dataclass(frozen=True)
class StreamingInterventionContext:
    """Information provided to intervention strategies for each micro-batch."""

    batch_id: int
    validation: ValidationResult
    dataset_id: str
    dataset_version: str


class StreamingInterventionStrategy(Protocol):
    """Decide whether a streaming pipeline should be interrupted."""

    def decide(self, context: StreamingInterventionContext) -> Optional[str]:
        """Return a reason to block the stream or ``None`` to continue."""


class NoOpStreamingInterventionStrategy:
    """Default strategy that never blocks the streaming pipeline."""

    def decide(self, context: StreamingInterventionContext) -> Optional[str]:
        return None


class StreamingObservationWriter:
    """Send streaming micro-batch observations to the data-quality service."""

    def __init__(
        self,
        *,
        contract: OpenDataContractStandard,
        expectation_plan: Sequence[Mapping[str, Any]],
        governance_service: GovernanceServiceClient,
        dataset_id: Optional[str],
        dataset_version: Optional[str],
        enforce: bool,
        checkpoint_location: Optional[str] = None,
        intervention: Optional[StreamingInterventionStrategy] = None,
        progress_callback: Optional[Callable[[Mapping[str, Any]], None]] = None,
    ) -> None:
        self.contract = contract
        self.expectation_plan = list(expectation_plan)
        self.governance_service = governance_service
        self.dataset_id = dataset_id or "unknown"
        self.dataset_version = dataset_version or "unknown"
        self.enforce = enforce
        self._validation: Optional[ValidationResult] = None
        self._latest_batch_id: Optional[int] = None
        self._active = False
        self._checkpoint_location = _derive_metrics_checkpoint(
            checkpoint_location,
            self.dataset_id,
            self.dataset_version,
        )
        default_name = f"dc43_metrics_{_safe_fs_name(self.dataset_id)}"
        self.query_name = f"{default_name}_{_safe_fs_name(self.dataset_version)}"
        self._intervention = intervention or NoOpStreamingInterventionStrategy()
        self._batches: List[Dict[str, Any]] = []
        self._progress_callback = progress_callback
        self._sink_queries: List[Any] = []

    @property
    def checkpoint_location(self) -> str:
        """Location used to checkpoint the metrics query."""
        return self._checkpoint_location

    @property
    def active(self) -> bool:
        """Whether the observation writer has already started its query."""
        return self._active

    def __getstate__(self) -> Dict[str, Any]:
        state = self.__dict__.copy()
        # Remove objects that cannot be pickled (e.g. Spark queries, callbacks)
        state['_sink_queries'] = []
        state['_progress_callback'] = None
        # self._validation might also contain issues if it's large, but typically it's a dataclass.
        return state

    def __setstate__(self, state: Dict[str, Any]) -> None:
        self.__dict__.update(state)

    def attach_validation(self, validation: ValidationResult) -> None:
        """Attach the validation object that should receive streaming metrics."""
        if self._validation is not None and self._validation is not validation:
            raise RuntimeError("StreamingObservationWriter already bound to a validation")

        self._validation = validation
        validation.merge_details(
            {
                "dataset_id": self.dataset_id,
                "dataset_version": self.dataset_version,
            }
        )

    def latest_validation(self) -> Optional[ValidationResult]:
        """Return the most recent validation produced by the observer."""
        return self._validation

    def streaming_batches(self) -> List[Mapping[str, Any]]:
        """Return the recorded micro-batch timeline."""
        return [dict(item) for item in self._batches]

    def _record_batch(
        self,
        *,
        batch_id: int,
        metrics: Mapping[str, Any] | None,
        row_count: int,
        status: str,
        timestamp: datetime,
        errors: Optional[Sequence[str]] = None,
        warnings: Optional[Sequence[str]] = None,
        intervention: Optional[str] = None,
    ) -> None:
        metrics_map = dict(metrics or {})
        violation_total = sum(
            int(value)
            for key, value in metrics_map.items()
            if key.startswith("violations.") and isinstance(value, (int, float))
        )
        entry: Dict[str, Any] = {
            "batch_id": batch_id,
            "timestamp": timestamp.isoformat(),
            "row_count": row_count,
            "violations": violation_total,
            "status": status,
        }
        if metrics_map:
            entry["metrics"] = metrics_map
        if errors:
            entry["errors"] = list(errors)
        if warnings:
            entry["warnings"] = list(warnings)
        if intervention:
            entry["intervention"] = intervention
        self._batches.append(entry)
        self._notify_progress({"type": "batch", **entry})

    def _notify_progress(self, event: Mapping[str, Any]) -> None:
        if self._progress_callback is None:
            return
        try:
            self._progress_callback(dict(event))
        except Exception:  # pragma: no cover - best effort progress hook
            logger.exception("Streaming progress callback failed")

    def watch_sink_query(self, query: Any) -> None:
        """Track a sink query so it can be stopped on enforcement failure."""
        if query not in self._sink_queries:
            self._sink_queries.append(query)

    def _stop_sink_queries(self) -> None:
        for query in list(self._sink_queries):
            try:
                stop = query.stop  # type: ignore[attr-defined]
            except AttributeError:
                stop = None
            try:
                if callable(stop):
                    stop()
            except Exception:  # pragma: no cover - best effort cleanup
                logger.exception("Failed to stop streaming sink query")

    def _merge_batch_details(
        self,
        result: ValidationResult,
        *,
        batch_id: int,
    ) -> None:
        details = {
            "dataset_id": self.dataset_id,
            "dataset_version": self.dataset_version,
            "streaming_batch_id": batch_id,
        }
        if result.metrics:
            details["streaming_metrics"] = dict(result.metrics)
        if self._batches:
            details["streaming_batches"] = [dict(item) for item in self._batches]
        result.merge_details(details)
        if self._validation is not None:
            validation = self._validation
            validation.ok = result.ok
            validation.errors = list(result.errors)
            validation.warnings = list(result.warnings)
            validation.metrics = dict(result.metrics)
            validation.schema = dict(result.schema)
            validation.status = result.status
            validation.reason = result.reason
            validation.merge_details(details)
            self._validation = validation
        else:
            self._validation = result

    def process_batch(self, batch_df: DataFrame, batch_id: int) -> ValidationResult:
        """Validate a micro-batch and update the attached validation."""
        timestamp = datetime.now(timezone.utc)
        schema, metrics = collect_observations(
            batch_df,
            self.contract,
            expectations=self.expectation_plan,
            collect_metrics=True,
        )
        row_count = metrics.get("row_count")
        if isinstance(row_count, (int, float)) and row_count <= 0:
            logger.debug(
                "Skipping empty streaming batch %s for %s@%s",
                batch_id,
                self.dataset_id,
                self.dataset_version,
            )
            self._latest_batch_id = batch_id
            validation = self._validation
            if validation is None:
                validation = ValidationResult(ok=True, errors=[], warnings=[])
            validation.merge_details(
                {
                    "dataset_id": self.dataset_id,
                    "dataset_version": self.dataset_version,
                    "streaming_batch_id": batch_id,
                }
            )
            self._validation = validation
            self._record_batch(
                batch_id=batch_id,
                metrics={},
                row_count=0,
                status="idle",
                timestamp=timestamp,
            )
            return validation

        cid, cver = contract_identity(self.contract)
        def _obs(): return ObservationPayload(metrics=dict(metrics or {}), schema=dict(schema or {}), reused=False)
        assessment = self.governance_service.evaluate_dataset(
            contract_id=cid,
            contract_version=cver,
            dataset_id=self.dataset_id,
            dataset_version=self.dataset_version,
            validation=None,
            observations=_obs,
            pipeline_context=None,
            operation="write",
        )
        result = assessment.validation or assessment.status
        if result is None:
            result = self._evaluate_without_service(schema=schema, metrics=metrics)
        self._latest_batch_id = batch_id
        status = "ok"
        if result.errors:
            status = "error"
        elif result.warnings:
            status = "warning"
        self._record_batch(
            batch_id=batch_id,
            metrics=result.metrics or metrics,
            row_count=int(metrics.get("row_count", 0) or 0),
            status=status,
            timestamp=timestamp,
            errors=result.errors if result.errors else None,
            warnings=result.warnings if result.warnings else None,
        )
        self._merge_batch_details(result, batch_id=batch_id)

        if self.enforce and not result.ok:
            self._stop_sink_queries()
            raise ValueError(
                "Streaming batch %s failed data-quality validation: %s"
                % (batch_id, result.errors)
            )

        decision = self._intervention.decide(
            StreamingInterventionContext(
                batch_id=batch_id,
                validation=result,
                dataset_id=self.dataset_id,
                dataset_version=self.dataset_version,
            )
        )
        if decision:
            if self._batches:
                self._batches[-1]["intervention"] = decision
            batches_payload = [dict(item) for item in self._batches]
            if self._validation is not None:
                self._validation.merge_details({"streaming_batches": batches_payload})
            result.merge_details({"streaming_batches": batches_payload})
            reason_details = {"streaming_intervention_reason": decision}
            if self._validation is not None:
                self._validation.merge_details(reason_details)
            result.merge_details(reason_details)
            self._notify_progress(
                {
                    "type": "intervention",
                    "batch_id": batch_id,
                    "reason": decision,
                }
            )
            self._stop_sink_queries()
            raise StreamingInterventionError(decision)

        return result

    def start(self, df: DataFrame, *, output_mode: str) -> "StreamingQuery":
        """Start the observation writer for ``df`` and return its query handle."""
        if self._active:
            raise RuntimeError("StreamingObservationWriter can only be started once")
        self._active = True

        def _run(batch_df: DataFrame, batch_id: int) -> None:
            self.process_batch(batch_df, batch_id)

        writer = df.writeStream.foreachBatch(_run).outputMode(output_mode)
        writer = writer.option("checkpointLocation", self.checkpoint_location)
        if self.query_name:
            writer = writer.queryName(self.query_name)
        query = writer.start()
        try:
            query_name = query.name  # type: ignore[attr-defined]
        except AttributeError:
            query_name = self.query_name
        try:
            query_id = query.id  # type: ignore[attr-defined]
        except AttributeError:
            query_id = ""
        self._notify_progress(
            {
                "type": "observer-started",
                "query_name": query_name,
                "id": query_id,
            }
        )
        return query

    def _evaluate_without_service(
        self,
        *,
        schema: Mapping[str, Mapping[str, Any]],
        metrics: Mapping[str, Any],
    ) -> ValidationResult:
        """Return a validation outcome without a dedicated DQ service."""
        violation_counts: Dict[str, float] = {}
        for key, value in metrics.items():
            if not isinstance(key, str) or not key.startswith("violations."):
                continue
            suffix = key.partition(".")[2]
            try:
                violation_counts[suffix] = float(value)
            except (TypeError, ValueError):
                continue

        errors: list[str] = []
        warnings: list[str] = []
        for descriptor in self.expectation_plan:
            if not isinstance(descriptor, Mapping):
                continue
            key = descriptor.get("key")
            if not isinstance(key, str):
                continue
            count = violation_counts.get(key, 0.0)
            if count <= 0:
                continue
            message = f"Expectation {key} reported {int(count)} violation(s)"
            if bool(descriptor.get("optional")):
                warnings.append(message)
            else:
                errors.append(message)

        ok = not errors
        status = "ok"
        if errors:
            status = "block"
        elif warnings:
            status = "warn"

        return ValidationResult(
            ok=ok,
            errors=errors,
            warnings=warnings,
            metrics=dict(metrics),
            schema=dict(schema),
            status=status,
        )



"""Provide pipeline stub fragments and project files for Spark runtimes."""

from __future__ import annotations

from textwrap import dedent
from typing import Callable, Mapping

from .. import PipelineFile, PipelineProject, PipelineStub, register_pipeline_stub


def _coerce_hint(hints: Mapping[str, object], key: str) -> str:
    value = hints.get(key)
    if value is None:
        return ""
    return str(value).strip()


def _spark_readme(hints: Mapping[str, object]) -> str:
    runtime = _coerce_hint(hints, "spark_runtime")
    workspace = _coerce_hint(hints, "spark_workspace_url")
    profile = _coerce_hint(hints, "spark_workspace_profile")
    cluster = _coerce_hint(hints, "spark_cluster")

    wizard_hints = ["Wizard hints captured during export:"]
    if runtime:
        wizard_hints.append(f"- Runtime: {runtime}")
    if workspace:
        wizard_hints.append(f"- Workspace URL: {workspace}")
    if profile:
        wizard_hints.append(f"- CLI profile: {profile}")
    if cluster:
        wizard_hints.append(f"- Cluster reference: {cluster}")
    if len(wizard_hints) == 1:
        wizard_hints.append("- (no runtime-specific hints were recorded)")

    return (
        dedent(
            """
            # Spark pipeline example

            This directory contains a full PySpark pipeline that uses the dc43
            integration helpers exported by the setup wizard. The `main.py`
            entrypoint shows how to:

            - load service backends via ``bootstrap_pipeline.load_backends``
            - build a Spark session with ``build_spark_context``
            - read governed product inputs with automatic casting and
              validation
            - apply a simple transformation
            - publish results through the data-product governance helpers

            Update the placeholder constants in ``main.py`` before running the
            pipeline and remove or replace the demo transformation with your
            business logic.
            """
        ).strip()
        + "\n\n"
        + "\n".join(wizard_hints)
        + "\n"
    )


def _spark_main_script() -> str:
    return dedent(
        '''
        #!/usr/bin/env python3
        """Run the Spark pipeline example generated by the dc43 setup wizard."""

        from __future__ import annotations

        from bootstrap_pipeline import build_spark_context, load_backends
        from spark_pipeline import io as spark_io
        from spark_pipeline import transformations as spark_transformations


        DATA_PRODUCT_ID = "replace-with-data-product-id"
        INPUT_PORT = "replace-with-input-port"
        OUTPUT_PORT = "replace-with-output-port"


        def main() -> None:
            suite = load_backends()
            context = build_spark_context(app_name="dc43-spark-example")
            spark = context.get("spark")
            if spark is None:
                raise RuntimeError(
                    "build_spark_context did not return a Spark session. "
                    "Install pyspark and ensure the runtime is available."
                )

            print("[spark] Spark session initialised:", spark)

            governance_service = context.get(
                "governance_service", suite.governance
            )

            source_df, status = spark_io.read_data_product_input(
                spark=spark,
                governance_service=governance_service,
                data_product=DATA_PRODUCT_ID,
                input_port=INPUT_PORT,
            )
            if status is not None:
                print(
                    "[data_quality] Input status:",
                    status.status,
                    status.reason or "",
                )

            enriched_df = spark_transformations.build_output_frame(source_df)

            write_validation = spark_io.write_data_product_output(
                dataframe=enriched_df,
                governance_service=governance_service,
                output_port=OUTPUT_PORT,
                data_product=DATA_PRODUCT_ID,
            )

            print(
                "[done] Example pipeline completed. Replace placeholder identifiers",
                "before running in production.",
            )


        if __name__ == "__main__":
            main()
        '''
    ).strip() + "\n"


def _spark_io_module() -> str:
    return dedent(
        '''
        """Read/write helpers used by the Spark pipeline example."""

        from __future__ import annotations

        from typing import Optional, Tuple

        from dc43_integrations.spark.io import (
            GovernanceSparkReadRequest,
            GovernanceSparkWriteRequest,
            read_with_governance,
            write_with_governance,
        )
        from dc43_integrations.spark.violation_strategy import SplitWriteViolationStrategy
        from dc43_service_clients.data_products import (
            DataProductInputBinding,
            DataProductOutputBinding,
        )
        from dc43_service_clients.data_quality import ValidationResult
        from dc43_service_clients.governance.client.interface import GovernanceServiceClient
        from dc43_service_clients.governance.models import GovernanceReadContext, GovernanceWriteContext
        from pyspark.sql import DataFrame, SparkSession


        def read_data_product_input(
            *,
            spark: SparkSession,
            governance_service: GovernanceServiceClient,
            data_product: str,
            input_port: str,
        ) -> Tuple[DataFrame, Optional[ValidationResult]]:
            """Load a product input port enforced through governance."""

            request = GovernanceSparkReadRequest(
                context=GovernanceReadContext(
                    input_binding=DataProductInputBinding(
                        data_product=data_product,
                        port_name=input_port,
                    )
                ),
            )
            df, status = read_with_governance(
                spark,
                request,
                governance_service=governance_service,
                enforce=True,
                auto_cast=True,
                return_status=True,
            )
            return df, status


        def write_data_product_output(
            *,
            dataframe: DataFrame,
            governance_service: GovernanceServiceClient,
            output_port: str,
            data_product: str,
        ) -> ValidationResult:
            """Persist pipeline results relying solely on governance."""

            request = GovernanceSparkWriteRequest(
                context=GovernanceWriteContext(
                    output_binding=DataProductOutputBinding(
                        data_product=data_product,
                        port_name=output_port,
                    ),
                ),
                mode="overwrite",
            )

            validation, reject_status = write_with_governance(
                df=dataframe,
                request=request,
                governance_service=governance_service,
                enforce=True,
                auto_cast=True,
                return_status=True,
                violation_strategy=SplitWriteViolationStrategy(
                    valid_suffix="valid",
                    reject_suffix="reject",
                ),
            )

            if reject_status is not None:
                print(
                    "[write] Reject status:",
                    reject_status.status,
                    reject_status.reason or "",
                )

            return validation
        '''
    ).strip() + "\n"


def _spark_transformations_module() -> str:
    return dedent(
        '''
        """Demo transformations applied by the Spark example pipeline."""

        from __future__ import annotations

        from pyspark.sql import DataFrame, functions as F


        def build_output_frame(source: DataFrame) -> DataFrame:
            """Return a demo dataset derived from ``source``.

            Replace this logic with your business-specific transformations once the
            scaffold is in place.
            """

            enriched = source.withColumn(
                "processed_at",
                F.current_timestamp(),
            )
            if "status" in source.columns:
                enriched = enriched.withColumn(
                    "is_valid",
                    F.when(F.col("status") == F.lit("valid"), F.lit(True)).otherwise(False),
                )
            return enriched
        '''
    ).strip() + "\n"


def _spark_project(hints: Mapping[str, object]) -> PipelineProject:
    return PipelineProject(
        root="spark_pipeline",
        entrypoint="main.py",
        files=(
            PipelineFile(path="README.md", content=_spark_readme(hints)),
            PipelineFile(
                path="__init__.py",
                content='"""Spark pipeline example package."""\n',
            ),
            PipelineFile(path="main.py", content=_spark_main_script(), executable=True),
            PipelineFile(path="io.py", content=_spark_io_module()),
            PipelineFile(path="transformations.py", content=_spark_transformations_module()),
        ),
    )


def _spark_pipeline_stub(
    *,
    hints: Mapping[str, object],
    flags: Mapping[str, bool],
    json_literal: Callable[[object | None], str],
) -> PipelineStub:
    runtime_hint = json_literal(hints.get("spark_runtime"))
    workspace_hint = json_literal(hints.get("spark_workspace_url"))
    profile_hint = json_literal(hints.get("spark_workspace_profile"))
    cluster_hint = json_literal(hints.get("spark_cluster"))
    data_product_id = json_literal("replace-with-data-product-id")
    input_port = json_literal("replace-with-input-port")
    output_port = json_literal("replace-with-output-port")

    main_lines = (
        "    if integration == 'spark':",
        "        context = build_spark_context(app_name=\"dc43-pipeline-example\")",
        "        spark = context.get('spark')",
        "        if spark is None:",
        "            raise RuntimeError('Spark session was not initialised by build_spark_context')",
        "        print(\"[spark] Spark session initialised:\", spark)",
        f"        runtime_hint = {runtime_hint}",
        "        if runtime_hint:",
        "            print(\"[spark] Runtime configured in setup:\", runtime_hint)",
        f"        workspace_hint = {workspace_hint}",
        "        if workspace_hint:",
        "            print(\"[spark] Workspace URL:\", workspace_hint)",
        f"        profile_hint = {profile_hint}",
        "        if profile_hint:",
        "            print(\"[spark] CLI profile:\", profile_hint)",
        f"        cluster_hint = {cluster_hint}",
        "        if cluster_hint:",
        "            print(\"[spark] Cluster reference:\", cluster_hint)",
        "        governance_service = context.get('governance_service', governance_service)",
        f"        data_product_id = {data_product_id}",
        f"        input_port = {input_port}",
        f"        output_port = {output_port}",
        "        source_df, input_status = spark_io.read_data_product_input(",
        "            spark=spark,",
        "            governance_service=governance_service,",
        "            data_product=data_product_id,",
        "            input_port=input_port,",
        "        )",
        "        if input_status:",
        "            print(\"[spark] Input status:\", input_status.status, input_status.reason or \"\")",
        "        enriched_df = spark_transformations.build_output_frame(source_df)",
        "        write_validation = spark_io.write_data_product_output(",
        "            dataframe=enriched_df,",
        "            governance_service=governance_service,",
        "            output_port=output_port,",
        "            data_product=data_product_id,",
        "        )",
        "        print(\"[spark] Output validation status:\", write_validation.status)",
    )

    tail_lines = (
        "    if integration == 'spark':",
        "        print(\"[spark] Explore examples/spark_pipeline for modular helpers.\")",
    )

    return PipelineStub(
        bootstrap_imports=("build_spark_context",),
        additional_imports=(
            "from spark_pipeline import io as spark_io",
            "from spark_pipeline import transformations as spark_transformations",
        ),
        main_lines=main_lines,
        tail_lines=tail_lines,
        project=_spark_project(hints),
    )


register_pipeline_stub("spark", _spark_pipeline_stub)


__all__ = ["_spark_pipeline_stub"]




from __future__ import annotations

import logging
from dataclasses import dataclass, field
from datetime import datetime, timezone
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Mapping,
    Optional,
    Sequence,
    Tuple,
)

from pyspark.sql import DataFrame, SparkSession
from open_data_contract_standard.model import OpenDataContractStandard, Server  # type: ignore

from dc43_service_backends.core.odcs import custom_properties_dict
from dc43_service_backends.core.versioning import SemVer

from dc43_integrations.spark.io.common import (
    dataset_id_from_ref,
    _promote_delta_path_to_table,
    _safe_fs_name,
)
from dc43_integrations.spark.io.resolution import (
    DatasetResolution,
    DatasetLocatorStrategy,
)

logger = logging.getLogger(__name__)


def _timestamp() -> str:
    """Return an ISO timestamp suitable for dataset versioning."""
    now = datetime.utcnow().replace(tzinfo=timezone.utc)
    return now.isoformat().replace("+00:00", "Z")


def _ref_from_contract(contract: OpenDataContractStandard) -> tuple[Optional[str], Optional[str]]:
    """Return ``(path, table)`` derived from the contract's first server."""
    if not contract.servers:
        return None, None
    server: Server = contract.servers[0]
    try:
        path = server.path  # type: ignore
    except AttributeError:
        path = None
    if path:
        return path, None
    # Build table name from catalog/schema/database/dataset parts when present
    try:
        last = server.dataset or server.database  # type: ignore
        parts = [server.catalog, server.schema_, last]  # type: ignore
        table = ".".join([p for p in parts if p]) if any(parts) else None
        return None, table
    except AttributeError:
        return None, None


@dataclass
class ContractFirstDatasetLocator:
    """Default locator that favours contract servers over provided hints."""

    clock: Callable[[], str] = _timestamp

    def _resolve_base(
        self,
        contract: Optional[OpenDataContractStandard],
        *,
        path: Optional[str],
        table: Optional[str],
        format: Optional[str],
        spark: Optional[SparkSession] = None,
    ) -> tuple[Optional[str], Optional[str], Optional[str], Optional[Server]]:
        server: Optional[Server] = None
        if contract and contract.servers:
            c_path, c_table = _ref_from_contract(contract)
            server = contract.servers[0]
            if server is not None:
                try:
                    c_format = server.format  # type: ignore[attr-defined]
                except AttributeError:
                    c_format = None
            else:
                c_format = None
            if c_path is not None:
                path = c_path
            if c_table is not None:
                table = c_table
            if c_format is not None and format is None:
                format = c_format
        path, table = _promote_delta_path_to_table(
            path=path,
            table=table,
            format=format,
            spark=spark,
        )
        return path, table, format, server

    def _resolution(
        self,
        contract: Optional[OpenDataContractStandard],
        *,
        path: Optional[str],
        table: Optional[str],
        format: Optional[str],
        include_timestamp: bool,
    ) -> DatasetResolution:
        dataset_id = contract.id if contract else dataset_id_from_ref(table=table, path=path)
        dataset_version = self.clock() if include_timestamp else None
        server_props: Optional[Dict[str, Any]] = None
        read_options: Optional[Dict[str, str]] = None
        write_options: Optional[Dict[str, str]] = None
        if contract and contract.servers:
            first = contract.servers[0]
            props = custom_properties_dict(first)
            if props:
                server_props = props
                versioning = props.get(ContractVersionLocator.VERSIONING_PROPERTY)
                if isinstance(versioning, Mapping):
                    read_map = versioning.get("readOptions")
                    if isinstance(read_map, Mapping):
                        read_options = {
                            str(k): str(v)
                            for k, v in read_map.items()
                            if v is not None
                        }
                    write_map = versioning.get("writeOptions")
                    if isinstance(write_map, Mapping):
                        write_options = {
                            str(k): str(v)
                            for k, v in write_map.items()
                            if v is not None
                        }
        return DatasetResolution(
            path=path,
            table=table,
            format=format,
            dataset_id=dataset_id,
            dataset_version=dataset_version,
            read_options=read_options,
            write_options=write_options,
            custom_properties=server_props,
            load_paths=None,
        )

    def for_read(
        self,
        *,
        contract: Optional[OpenDataContractStandard],
        spark: SparkSession,
        format: Optional[str],
        path: Optional[str],
        table: Optional[str],
    ) -> DatasetResolution:
        path, table, format, _ = self._resolve_base(
            contract,
            path=path,
            table=table,
            format=format,
            spark=spark,
        )
        return self._resolution(
            contract,
            path=path,
            table=table,
            format=format,
            include_timestamp=False,
        )

    def for_write(
        self,
        *,
        contract: Optional[OpenDataContractStandard],
        df: DataFrame,
        format: Optional[str],
        path: Optional[str],
        table: Optional[str],
    ) -> DatasetResolution:
        path, table, format, _ = self._resolve_base(
            contract,
            path=path,
            table=table,
            format=format,
            spark=None,
        )
        return self._resolution(
            contract,
            path=path,
            table=table,
            format=format,
            include_timestamp=True,
        )


@dataclass
class StaticDatasetLocator:
    """Locator overriding specific fields while delegating to a base strategy."""

    dataset_id: Optional[str] = None
    dataset_version: Optional[str] = None
    path: Optional[str] = None
    table: Optional[str] = None
    format: Optional[str] = None
    base: DatasetLocatorStrategy = field(default_factory=ContractFirstDatasetLocator)

    def _merge(self, resolution: DatasetResolution) -> DatasetResolution:
        return DatasetResolution(
            path=self.path or resolution.path,
            table=self.table or resolution.table,
            format=self.format or resolution.format,
            dataset_id=self.dataset_id or resolution.dataset_id,
            dataset_version=self.dataset_version or resolution.dataset_version,
            read_options=dict(resolution.read_options or {}),
            write_options=dict(resolution.write_options or {}),
            custom_properties=resolution.custom_properties,
            load_paths=list(resolution.load_paths or []),
        )

    def for_read(
        self,
        *,
        contract: Optional[OpenDataContractStandard],
        spark: SparkSession,
        format: Optional[str],
        path: Optional[str],
        table: Optional[str],
    ) -> DatasetResolution:
        base_resolution = self.base.for_read(
            contract=contract,
            spark=spark,
            format=format,
            path=path,
            table=table,
        )
        return self._merge(base_resolution)

    def for_write(
        self,
        *,
        contract: Optional[OpenDataContractStandard],
        df: DataFrame,
        format: Optional[str],
        path: Optional[str],
        table: Optional[str],
    ) -> DatasetResolution:
        base_resolution = self.base.for_write(
            contract=contract,
            df=df,
            format=format,
            path=path,
            table=table,
        )
        return self._merge(base_resolution)


@dataclass
class ContractVersionLocator:
    """Locator that appends a version directory or time-travel hint."""

    dataset_version: str
    dataset_id: Optional[str] = None
    subpath: Optional[str] = None
    base: DatasetLocatorStrategy = field(default_factory=ContractFirstDatasetLocator)

    VERSIONING_PROPERTY = "dc43.core.versioning"

    @staticmethod
    def _version_key(value: str) -> tuple[int, Tuple[int, int, int] | float | str, str]:
        candidate = value
        if candidate.endswith("Z"):
            candidate = candidate[:-1] + "+00:00"
        try:
            dt = datetime.fromisoformat(candidate)
            return (0, dt.timestamp(), value)
        except ValueError:
            pass
        try:
            parsed = SemVer.parse(value)
            return (1, (parsed.major, parsed.minor, parsed.patch), value)
        except ValueError:
            return (2, value, value)

    @staticmethod
    def _render_template(template: str, *, version_value: str, safe_value: str) -> str:
        return (
            template.replace("{version}", version_value)
            .replace("{safeVersion}", safe_value)
        )

    @staticmethod
    def _folder_version_value(path: Path) -> str:
        marker = path / ".dc43_version"
        if marker.exists():
            try:
                text = marker.read_text().strip()
            except OSError:
                text = ""
            if text:
                return text
        return path.name

    @classmethod
    def _versioning_config(cls, resolution: DatasetResolution) -> Optional[Mapping[str, Any]]:
        props = resolution.custom_properties or {}
        value = props.get(cls.VERSIONING_PROPERTY)
        if isinstance(value, Mapping):
            return value
        return None

    @classmethod
    def _expand_versioning_paths(
        cls,
        resolution: DatasetResolution,
        *,
        base_path: Optional[str],
        dataset_version: Optional[str],
    ) -> tuple[Optional[List[str]], Dict[str, str]]:
        config = cls._versioning_config(resolution)
        if not config or not base_path or not dataset_version:
            return None, {}

        base = Path(base_path)
        base_dir = base.parent if base.suffix else base
        if not base_dir.exists():
            return None, {}

        include_prior = bool(config.get("includePriorVersions"))
        folder_template = str(config.get("subfolder", "{version}"))
        file_pattern = config.get("filePattern")
        if file_pattern is not None:
            file_pattern = str(file_pattern)
        elif base.suffix:
            file_pattern = base.name

        dataset_version_normalised = dataset_version
        lower = dataset_version.lower()
        entries: List[tuple[str, str]] = []
        try:
            for entry in base_dir.iterdir():
                if not entry.is_dir():
                    continue
                display = cls._folder_version_value(entry)
                entries.append((display, entry.name))
        except FileNotFoundError:
            return None, {}
        if not entries:
            return None, {}
        entries.sort(key=lambda item: cls._version_key(item[0]))

        selected: List[tuple[str, str]] = []
        if lower == "latest":
            alias_key = None
            alias_path = base_dir / dataset_version_normalised
            if alias_path.exists():
                try:
                    resolved_alias = alias_path.resolve()
                except OSError:
                    resolved_alias = alias_path
                if resolved_alias.is_dir():
                    alias_display = cls._folder_version_value(resolved_alias)
                    alias_key = cls._version_key(alias_display)

            if include_prior:
                if alias_key is not None:
                    selected = [
                        entry for entry in entries if cls._version_key(entry[0]) <= alias_key
                    ]
                else:
                    selected = entries
            elif entries:
                if alias_key is not None:
                    selected = [
                        entry for entry in entries if cls._version_key(entry[0]) == alias_key
                    ]
                    if not selected and entries:
                        selected = [entries[-1]]
                else:
                    selected = [entries[-1]]
        else:
            target_key = cls._version_key(dataset_version_normalised)
            eligible = [entry for entry in entries if cls._version_key(entry[0]) <= target_key]
            alias_like = "__" in dataset_version_normalised
            effective_include_prior = include_prior and not alias_like
            if effective_include_prior:
                selected = eligible
            else:
                exact = next((entry for entry in entries if entry[0] == dataset_version_normalised), None)
                if exact:
                    selected = [exact]
                else:
                    safe_candidate = _safe_fs_name(dataset_version_normalised)
                    fallback = next((entry for entry in entries if entry[1] == safe_candidate), None)
                    if fallback:
                        selected = [fallback]
                    elif eligible:
                        selected = [eligible[-1]]

        if not selected:
            candidate_path = base_dir / dataset_version_normalised
            if candidate_path.exists():
                selected = [(dataset_version_normalised, candidate_path.name)]
            else:
                return None, {}

        resolved_paths: List[str] = []
        for display_value, folder_name in selected:
            rendered_folder = cls._render_template(
                folder_template,
                version_value=display_value,
                safe_value=folder_name,
            )
            root = base_dir / rendered_folder if rendered_folder else base_dir
            if not root.exists():
                fallback_root = base_dir / folder_name
                if fallback_root.exists():
                    root = fallback_root
            if file_pattern:
                pattern = cls._render_template(
                    file_pattern,
                    version_value=display_value,
                    safe_value=folder_name,
                )
                matches = list(root.glob(pattern))
                if matches:
                    resolved_paths.extend(str(path) for path in matches)
            else:
                if root.exists():
                    resolved_paths.append(str(root))

        read_opts: Dict[str, str] = {}
        extra_read = config.get("readOptions")
        if isinstance(extra_read, Mapping):
            for k, v in extra_read.items():
                if isinstance(v, bool):
                    read_opts[str(k)] = str(v).lower()
                else:
                    read_opts[str(k)] = str(v)

        return (resolved_paths or None), read_opts

    def _resolve_path(self, resolution: DatasetResolution) -> Optional[str]:
        path = resolution.path
        if not path:
            return None

        fmt = (resolution.format or "").lower()
        if fmt == "delta":
            return path

        base = Path(path)
        safe_component: Optional[str] = None
        if self.dataset_version:
            candidate = _safe_fs_name(self.dataset_version)
            if candidate and candidate != self.dataset_version:
                safe_component = candidate

        if base.suffix:
            version_component = self.dataset_version
            parent = base.parent / base.stem
            if safe_component and version_component:
                preferred_dir = parent / version_component
                if not preferred_dir.exists():
                    version_component = safe_component
            elif safe_component and not version_component:
                version_component = safe_component

            folder = parent / version_component if version_component else parent
            if self.subpath:
                folder = folder / self.subpath
            target = folder / base.name
            return str(target)

        version_component = self.dataset_version
        if safe_component and version_component:
            preferred_dir = base / version_component
            if not preferred_dir.exists():
                version_component = safe_component
        elif safe_component and not version_component:
            version_component = safe_component

        folder = base / version_component if version_component else base
        if self.subpath:
            folder = folder / self.subpath
        return str(folder)

    @staticmethod
    def _delta_time_travel_option(dataset_version: Optional[str]) -> Optional[tuple[str, str]]:
        if not dataset_version:
            return None

        version = dataset_version.strip()
        if not version or version.lower() == "latest":
            return None

        if version.isdigit():
            return "versionAsOf", version

        candidate = version
        if candidate.endswith("Z"):
            candidate = candidate[:-1] + "+00:00"
        try:
            datetime.fromisoformat(candidate)
        except ValueError:
            return None
        return "timestampAsOf", version

    def _merge(
        self,
        contract: Optional[OpenDataContractStandard],
        resolution: DatasetResolution,
    ) -> DatasetResolution:
        resolved_path = self._resolve_path(resolution)
        dataset_id = self.dataset_id or resolution.dataset_id
        if dataset_id is None and contract is not None:
            dataset_id = contract.id
        read_options = dict(resolution.read_options or {})
        write_options = dict(resolution.write_options or {})
        load_paths = list(resolution.load_paths or [])
        base_path_hint = resolution.path
        version_paths, extra_read_options = self._expand_versioning_paths(
            resolution,
            base_path=base_path_hint,
            dataset_version=self.dataset_version,
        )
        if version_paths:
            load_paths = version_paths
            resolved_path = base_path_hint or resolved_path
        if extra_read_options:
            read_options.update(extra_read_options)
        if (resolution.format or "").lower() == "delta":
            option = self._delta_time_travel_option(self.dataset_version)
            if option:
                read_options.setdefault(*option)
        return DatasetResolution(
            path=resolved_path or resolution.path,
            table=resolution.table,
            format=resolution.format,
            dataset_id=dataset_id,
            dataset_version=self.dataset_version,
            read_options=read_options or None,
            write_options=write_options or None,
            custom_properties=resolution.custom_properties,
            load_paths=load_paths or None,
        )

    def for_read(
        self,
        *,
        contract: Optional[OpenDataContractStandard],
        spark: SparkSession,
        format: Optional[str],
        path: Optional[str],
        table: Optional[str],
    ) -> DatasetResolution:
        base_resolution = self.base.for_read(
            contract=contract,
            spark=spark,
            format=format,
            path=path,
            table=table,
        )
        return self._merge(contract, base_resolution)

    def for_write(
        self,
        *,
        contract: Optional[OpenDataContractStandard],
        df: DataFrame,
        format: Optional[str],
        path: Optional[str],
        table: Optional[str],
    ) -> DatasetResolution:
        base_resolution = self.base.for_write(
            contract=contract,
            df=df,
            format=format,
            path=path,
            table=table,
        )
        return self._merge(contract, base_resolution)



from __future__ import annotations

import copy
import logging
import warnings
from dataclasses import is_dataclass, replace
from pathlib import Path
from typing import (
    Any,
    Dict,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TYPE_CHECKING,
)

from dc43_service_backends.core.versioning import SemVer, version_key
from open_data_contract_standard.model import OpenDataContractStandard, Server  # type: ignore

if TYPE_CHECKING:
    from dc43_service_clients.contracts.client.interface import ContractServiceClient
    from dc43_service_clients.data_products import (
        DataProductServiceClient,
        OpenDataProductStandard,
        DataProductInputBinding,
    )
    from dc43_service_clients.governance.client.interface import GovernanceServiceClient
    from dc43_integrations.spark.io.strategies import SupportsDataProductStatusPolicy, SupportsContractStatusValidation, SupportsDataProductStatusValidation

logger = logging.getLogger(__name__)


def _paths_compatible(provided: str, contract_path: str) -> bool:
    """Return ``True`` when ``provided`` is consistent with ``contract_path``."""

    try:
        actual = Path(provided).resolve()
        expected = Path(contract_path).resolve()
    except OSError:
        return False

    if actual == expected:
        return True

    base = expected.parent / expected.stem if expected.suffix else expected
    if actual == base:
        return True

    return base in actual.parents


def _select_version(versions: list[str], minimum: str) -> str:
    """Return the highest version satisfying ``>= minimum``."""

    try:
        base = SemVer.parse(minimum)
    except ValueError as exc:  # pragma: no cover - defensive guard
        raise ValueError(f"Invalid minimum version: {minimum}") from exc

    best: tuple[int, int, int] | None = None
    best_value: Optional[str] = None
    for candidate in versions:
        try:
            parsed = SemVer.parse(candidate)
        except ValueError:
            # Fallback to string comparison when candidate matches exactly.
            if candidate == minimum:
                return candidate
            continue
        key = (parsed.major, parsed.minor, parsed.patch)
        if key < (base.major, base.minor, base.patch):
            continue
        if best is None or key > best:
            best = key
            best_value = candidate
    if best_value is None:
        raise ValueError(f"No versions found satisfying >= {minimum}")
    return best_value


def _resolve_contract(
    *,
    contract_id: str,
    expected_version: Optional[str],
    service: ContractServiceClient | None,
    governance: GovernanceServiceClient | None,
) -> OpenDataContractStandard:
    """Fetch a contract from the configured service respecting version hints."""

    if service is None and governance is None:
        raise ValueError(
            "A contract service or governance service is required when contract_id is provided",
        )

    def _resolve_version(candidate: Optional[str]) -> str:
        if not candidate:
            if governance is not None:
                contract = governance.latest_contract(contract_id=contract_id)
            else:
                contract = service.latest(contract_id)  # type: ignore[union-attr]
            if contract is None:
                raise ValueError(f"No versions available for contract {contract_id}")
            return contract.version

        if candidate.startswith("=="):
            return candidate[2:]

        if candidate.startswith(">="):
            base = candidate[2:]
            versions: Sequence[str]
            if governance is not None:
                versions = tuple(governance.list_contract_versions(contract_id=contract_id))
            else:
                versions = tuple(service.list_versions(contract_id))  # type: ignore[union-attr]
            selected = _select_version(list(versions), base)
            return selected

        return candidate

    version = _resolve_version(expected_version)

    if governance is not None:
        return governance.get_contract(contract_id=contract_id, contract_version=version)

    return service.get(contract_id, version)  # type: ignore[union-attr]


def _enforce_contract_status(
    *,
    handler: object,
    contract: OpenDataContractStandard,
    enforce: bool,
    operation: str,
) -> None:
    """Apply a contract status policy defined by ``handler``."""

    # Using duck typing or Protocol check if imported
    if hasattr(handler, "validate_contract_status") and callable(handler.validate_contract_status): # type: ignore
         handler.validate_contract_status( # type: ignore
            contract=contract,
            enforce=enforce,
            operation=operation,
        )
         return

    _validate_contract_status(
        contract=contract,
        enforce=enforce,
        operation=operation,
    )


def _validate_contract_status(
    *,
    contract: OpenDataContractStandard,
    enforce: bool,
    operation: str,
    allowed_statuses: Iterable[str] | None = None,
    allow_missing: bool = True,
    case_insensitive: bool = True,
    failure_message: str | None = None,
) -> None:
    """Check the contract status against an allowed set."""

    raw_status = contract.status
    if raw_status is None:
        if allow_missing:
            return
        status_value = ""
    else:
        status_value = str(raw_status).strip()
        if not status_value and allow_missing:
            return

    if not status_value:
        message = (
            failure_message
            or "Contract {contract_id}:{contract_version} status {status!r} "
            "is not allowed for {operation} operations"
        ).format(
            contract_id=str(contract.id or ""),
            contract_version=str(contract.version or ""),
            status=status_value,
            operation=operation,
        )
        if enforce:
            raise ValueError(message)
        logger.warning(message)
        return

    options = allowed_statuses or ("active",)
    allowed = {status.lower() if case_insensitive else status for status in options}
    candidate = status_value.lower() if case_insensitive else status_value
    if candidate in allowed:
        return

    message = (
        failure_message
        or "Contract {contract_id}:{contract_version} status {status!r} "
        "is not allowed for {operation} operations"
    ).format(
        contract_id=str(contract.id or ""),
        contract_version=str(contract.version or ""),
        status=status_value,
        operation=operation,
    )
    if enforce:
        raise ValueError(message)
    logger.warning(message)


def _normalise_version_spec(spec: Optional[str]) -> Optional[str]:
    """Return a normalised version constraint or ``None`` when unset."""

    if spec is None:
        return None
    value = str(spec).strip()
    if not value:
        return None
    if value.startswith("=="):
        return value[2:].strip() or None
    return value


def _check_contract_version(
    expected: Optional[str],
    actual: Optional[str],
) -> None:
    """Raise ValueError if actual does not satisfy the expected version constraint."""
    if expected is None or not expected.strip():
        return
    requirement = expected.strip()

    if not actual:
        raise ValueError(f"Actual contract version is missing; expected {expected}")

    if requirement.startswith("=="):
        target = requirement[2:].strip()
        if actual != target:
            raise ValueError(f"Contract version {actual} does not satisfy {expected}")
        return

    if requirement.startswith(">="):
        target = requirement[2:].strip()
        if not target:
            return
        
        try:
            if version_key(actual) < version_key(target):
                raise ValueError(f"Contract version {actual} does not satisfy {expected}")
        except Exception as exc:
            raise ValueError(f"Unable to compare versions {actual!r} and {target!r}: {exc}") from exc
        return

    if actual != requirement:
        raise ValueError(f"Contract version {actual} does not satisfy {expected}")


def _check_data_product_version(
    *,
    expected: Optional[str],
    actual: Optional[str],
    data_product_id: str,
    subject: str,
    enforce: bool,
) -> bool:
    """Return ``True`` when ``actual`` satisfies the optional ``expected`` constraint."""

    if expected is None or not expected.strip():
        return True
    if not actual:
        message = (
            f"{subject} version for data product {data_product_id} is unknown; expected {expected}"
        )
        if enforce:
            raise ValueError(message)
        logger.warning(message)
        return False

    requirement = expected.strip()
    if requirement.startswith("=="):
        target = requirement[2:].strip()
        if actual != target:
            message = (
                f"{subject} version {actual} does not satisfy {expected} for data product {data_product_id}"
            )
            if enforce:
                raise ValueError(message)
            logger.warning(message)
            return False
        return True
    if requirement.startswith(">="):
        target = requirement[2:].strip()
        if not target:
            return True
        try:
            if version_key(actual) < version_key(target):
                message = (
                    f"{subject} version {actual} does not satisfy {expected} for data product {data_product_id}"
                )
                if enforce:
                    raise ValueError(message)
                logger.warning(message)
                return False
        except Exception as exc:  # pragma: no cover - defensive against malformed versions
            message = (
                f"Unable to compare versions {actual!r} and {target!r} for data product {data_product_id}: {exc}"
            )
            if enforce:
                raise ValueError(message) from exc
            logger.warning(message)
            return False
        return True
    if actual != requirement:
        message = (
            f"{subject} version {actual} does not satisfy {expected} for data product {data_product_id}"
        )
        if enforce:
            raise ValueError(message)
        logger.warning(message)
        return False
    return True


def _validate_data_product_status(
    *,
    data_product: OpenDataProductStandard,
    enforce: bool,
    operation: str,
    allowed_statuses: Iterable[str] | None = None,
    allow_missing: bool = True,
    case_insensitive: bool = True,
    failure_message: str | None = None,
) -> None:
    """Check the data product status against an allowed set."""

    raw_status = data_product.status
    product_id = str(data_product.id or "")
    product_version = str(data_product.version or "")
    if raw_status is None:
        if allow_missing:
            return
        status_value = ""
    else:
        status_value = str(raw_status).strip()
        if not status_value and allow_missing:
            return

    if not status_value:
        message = (
            failure_message
            or "Data product {data_product_id}@{data_product_version} status {status!r} "
            "is not allowed for {operation} operations"
        ).format(
            data_product_id=product_id,
            data_product_version=product_version,
            status=status_value,
            operation=operation,
        )
        if enforce:
            raise ValueError(message)
        logger.warning(message)
        return

    options = allowed_statuses or ("active",)
    allowed = {status.lower() if case_insensitive else status for status in options}
    candidate = status_value.lower() if case_insensitive else status_value
    if candidate in allowed:
        return

    message = (
        failure_message
        or "Data product {data_product_id}@{data_product_version} status {status!r} "
        "is not allowed for {operation} operations"
    ).format(
        data_product_id=product_id,
        data_product_version=product_version,
        status=status_value,
        operation=operation,
    )
    if enforce:
        raise ValueError(message)
    logger.warning(message)


def _clone_status_handler(handler: object, overrides: Mapping[str, Any]) -> object:
    """Return ``handler`` updated with ``overrides`` without mutating the input."""

    if not overrides:
        return handler
    if is_dataclass(handler):
        try:
            return replace(handler, **overrides) # type: ignore
        except TypeError:
            pass
    try:
        clone = copy.copy(handler)
    except Exception:  # pragma: no cover - fallback when cloning fails
        clone = handler
    for key, value in overrides.items():
        try:
            object.__getattribute__(clone, key)
        except AttributeError:
            continue
        setattr(clone, key, value)
    return clone


def _apply_plan_data_product_policy(
    handler: object,
    plan: Any | None,
    *,
    default_enforce: bool,
) -> tuple[object, bool]:
    """Return a handler/enforcement pair honouring plan-defined policies."""

    if plan is None:
        return handler, default_enforce

    overrides: Dict[str, Any] = {}
    policy = getattr(plan, "policy", None)
    source = policy if policy is not None else plan

    try:
        allowed_statuses = source.allowed_data_product_statuses  # type: ignore[attr-defined]
    except AttributeError:
        allowed_statuses = None
    if (
        allowed_statuses is not None
        and hasattr(handler, "allowed_data_product_statuses") # Duck type
    ):
        overrides["allowed_data_product_statuses"] = tuple(allowed_statuses)

    try:
        allow_missing = source.allow_missing_data_product_status  # type: ignore[attr-defined]
    except AttributeError:
        allow_missing = None
    if (
        allow_missing is not None
        and hasattr(handler, "allow_missing_data_product_status") # Duck type
    ):
        overrides["allow_missing_data_product_status"] = bool(allow_missing)

    try:
        case_insensitive = source.data_product_status_case_insensitive  # type: ignore[attr-defined]
    except AttributeError:
        case_insensitive = None
    if (
        case_insensitive is not None
        and hasattr(handler, "data_product_status_case_insensitive") # Duck type
    ):
        overrides["data_product_status_case_insensitive"] = bool(case_insensitive)

    try:
        failure_message = source.data_product_status_failure_message  # type: ignore[attr-defined]
    except AttributeError:
        failure_message = None
    if (
        failure_message is not None
        and hasattr(handler, "data_product_status_failure_message") # Duck type
    ):
        overrides["data_product_status_failure_message"] = failure_message

    handler = _clone_status_handler(handler, overrides)

    try:
        enforce_override = source.enforce_data_product_status  # type: ignore[attr-defined]
    except AttributeError:
        enforce_override = None
    if enforce_override is None:
        return handler, default_enforce
    return handler, bool(enforce_override)


def _enforce_data_product_status(
    *,
    handler: object,
    data_product: OpenDataProductStandard,
    enforce: bool,
    operation: str,
) -> None:
    """Apply a data product status policy defined by ``handler``."""

    if hasattr(handler, "validate_data_product_status") and callable(handler.validate_data_product_status): # type: ignore
        handler.validate_data_product_status( # type: ignore
            data_product=data_product,
            enforce=enforce,
            operation=operation,
        )
        return

    _validate_data_product_status(
        data_product=data_product,
        enforce=enforce,
        operation=operation,
    )


def _select_data_product(
    *,
    service: DataProductServiceClient,
    data_product_id: str,
    version_spec: Optional[str],
    handler: object,
    enforce: bool,
    operation: str,
    status_enforce: Optional[bool] = None,
) -> Optional[OpenDataProductStandard]:
    """Return a data product respecting ``handler`` status policies."""

    requirement = version_spec.strip() if isinstance(version_spec, str) else ""
    direct_version = None
    policy_enforce = enforce if status_enforce is None else status_enforce
    if requirement and not requirement.startswith(">="):
        direct_version = _normalise_version_spec(requirement)
    if direct_version:
        try:
            product = service.get(data_product_id, direct_version)
        except Exception:
            if enforce:
                raise
            logger.warning(
                "Data product %s version %s could not be retrieved",
                data_product_id,
                direct_version,
            )
            return None
        _enforce_data_product_status(
            handler=handler,
            data_product=product,
            enforce=policy_enforce,
            operation=operation,
        )
        if not _check_data_product_version(
            expected=requirement,
            actual=product.version,
            data_product_id=data_product_id,
            subject="Data product",
            enforce=enforce,
        ):
            return None
        return product

    latest: Optional[OpenDataProductStandard] = None
    try:
        latest = service.latest(data_product_id)
    except Exception:  # pragma: no cover - defensive logging
        logger.exception("Failed to resolve latest data product %s", data_product_id)

    candidates: list[tuple[Optional[str], Optional[OpenDataProductStandard]]] = []
    if latest is not None:
        candidates.append((latest.version, latest))

    versions: Iterable[str] = ()
    try:
        versions = service.list_versions(data_product_id)
    except Exception:  # pragma: no cover - defensive logging
        logger.exception("Failed to list versions for data product %s", data_product_id)

    seen_versions: set[str] = set()
    if latest and latest.version:
        seen_versions.add(latest.version)

    sorted_versions = sorted(
        (version for version in versions if version),
        key=version_key,
        reverse=True,
    )
    for version in sorted_versions:
        if version in seen_versions:
            continue
        seen_versions.add(version)
        candidates.append((version, None))

    errors: list[str] = []
    for version, product in candidates:
        candidate = product
        if candidate is None and version:
            try:
                candidate = service.get(data_product_id, version)
            except Exception:
                logger.exception(
                    "Failed to load data product %s version %s", data_product_id, version
                )
                continue
        if candidate is None:
            continue
        try:
            _enforce_data_product_status(
                handler=handler,
                data_product=candidate,
                enforce=policy_enforce,
                operation=operation,
            )
        except ValueError as exc:
            errors.append(str(exc))
            continue
        if requirement:
            matches = _check_data_product_version(
                expected=requirement,
                actual=candidate.version,
                data_product_id=data_product_id,
                subject="Data product",
                enforce=enforce,
            )
            if not matches:
                continue
        return candidate

    if errors:
        message = (
            f"Data product {data_product_id} does not have an allowed version for {operation} operations"
        )
        if enforce:
            detail = "; ".join(dict.fromkeys(errors))
            raise ValueError(f"{message}: {detail}")
        logger.warning("%s: %s", message, "; ".join(dict.fromkeys(errors)))
        return None

    if requirement:
        message = (
            f"Data product {data_product_id} has no versions available for {operation} operations"
        )
        if enforce:
            raise ValueError(message)
        logger.warning(message)
    return None


def _load_binding_product_version(
    *,
    service: DataProductServiceClient,
    data_product_id: str,
    version_spec: Optional[str],
    enforce: bool,
    operation: str,
) -> tuple[Optional[OpenDataProductStandard], bool]:
    """Return the exact product version requested by a binding when available."""

    requirement = _normalise_version_spec(version_spec)
    if not requirement or requirement.startswith(">="):
        return None, False
    try:
        product = service.get(data_product_id, requirement)
    except Exception as exc:
        message = (
            f"Data product {data_product_id} version {requirement} is not available "
            f"for {operation} registration"
        )
        if enforce:
            raise ValueError(message) from exc
        logger.warning(message)
        return None, False
    return product, True



"""Testing helpers for synthesising contract-aligned Spark datasets."""

from __future__ import annotations

import re
from datetime import datetime, timedelta
from decimal import Decimal
from typing import Callable, Iterable, List, Mapping, Sequence, Tuple

from faker import Faker
from open_data_contract_standard.model import (  # type: ignore
    OpenDataContractStandard,
    SchemaProperty,
)
from pyspark.sql import DataFrame, SparkSession
from pyspark.sql import types as T

from dc43_integrations.spark.data_quality import spark_type_name
from dc43_integrations.spark.validation import apply_contract


Generator = Callable[[Faker, SchemaProperty], object]

_DECIMAL_PATTERN = re.compile(r"decimal\s*\((\d+)\s*,\s*(\d+)\)", re.IGNORECASE)
_CHAR_PATTERN = re.compile(r"(?:var)?char\s*\((\d+)\)", re.IGNORECASE)


def _list_properties(contract: OpenDataContractStandard) -> List[SchemaProperty]:
    props: List[SchemaProperty] = []
    schema_objects = getattr(contract, "schema_", None) or getattr(contract, "schema", None) or []
    for obj in schema_objects:
        properties = getattr(obj, "properties", None) or []
        for prop in properties:
            if isinstance(prop, SchemaProperty):
                props.append(prop)
            elif isinstance(prop, Mapping):
                try:
                    props.append(SchemaProperty.model_validate(dict(prop)))
                except Exception:
                    continue
    return props


_SPARK_TYPES: dict[str, T.DataType] = {
    "string": T.StringType(),
    "bigint": T.LongType(),
    "long": T.LongType(),
    "int": T.IntegerType(),
    "integer": T.IntegerType(),
    "smallint": T.ShortType(),
    "tinyint": T.ByteType(),
    "double": T.DoubleType(),
    "float": T.FloatType(),
    "boolean": T.BooleanType(),
    "bool": T.BooleanType(),
    "date": T.DateType(),
    "timestamp": T.TimestampType(),
    "binary": T.BinaryType(),
}


def _enum_values(field: SchemaProperty) -> Sequence[object]:
    values: List[object] = []
    for rule in getattr(field, "quality", None) or []:
        rule_name = str(getattr(rule, "rule", "")).lower()
        if rule_name == "enum":
            raw = getattr(rule, "mustBe", None) or getattr(rule, "values", None)
            if raw is None:
                continue
            if isinstance(raw, Iterable) and not isinstance(raw, (str, bytes, bytearray)):
                values.extend(raw)
            else:
                values.append(raw)
    return values


def _string_length_hint(type_hint: str) -> int | None:
    match = _CHAR_PATTERN.search(type_hint)
    if match:
        try:
            return int(match.group(1))
        except ValueError:
            return None
    return None


def _string_generator(fake: Faker, field: SchemaProperty) -> str:
    choices = _enum_values(field)
    if choices:
        return str(fake.random_element(choices))
    type_hint = str(field.physicalType or field.logicalType or "").lower()
    max_len = _string_length_hint(type_hint)
    if max_len is not None:
        return fake.pystr(min_chars=1, max_chars=max(1, min(max_len, 32)))
    return fake.pystr(min_chars=5, max_chars=24)


def _integer_generator(bounds: Tuple[int, int]) -> Generator:
    def _gen(fake: Faker, _: SchemaProperty) -> int:
        return int(fake.random_int(min=bounds[0], max=bounds[1]))

    return _gen


def _floating_generator(right_digits: int = 6) -> Generator:
    def _gen(fake: Faker, _: SchemaProperty) -> float:
        return float(fake.pyfloat(left_digits=6, right_digits=right_digits))

    return _gen


def _boolean_generator(fake: Faker, _: SchemaProperty) -> bool:
    return bool(fake.pybool())


_REFERENCE_DATETIME = datetime(2025, 1, 1, 0, 0, 0)


def _date_generator(fake: Faker, _: SchemaProperty):
    start = (_REFERENCE_DATETIME - timedelta(days=30)).date()
    end = _REFERENCE_DATETIME.date()
    return fake.date_between(start_date=start, end_date=end)


def _timestamp_generator(fake: Faker, _: SchemaProperty):
    start = _REFERENCE_DATETIME - timedelta(days=30)
    end = _REFERENCE_DATETIME
    return fake.date_time_between_dates(datetime_start=start, datetime_end=end)


def _binary_generator(fake: Faker, _: SchemaProperty) -> bytes:
    return bytes(fake.binary(length=16))


def _decimal_spec(type_hint: str) -> tuple[T.DecimalType, Generator]:
    match = _DECIMAL_PATTERN.search(type_hint)
    if match:
        precision = int(match.group(1))
        scale = int(match.group(2))
    else:
        precision, scale = 18, 6
    precision = max(1, precision)
    scale = max(0, min(scale, precision))
    dtype = T.DecimalType(precision, scale)

    left_digits = max(1, precision - scale)

    def _gen(fake: Faker, _: SchemaProperty) -> Decimal:
        return fake.pydecimal(left_digits=left_digits, right_digits=scale, positive=False)

    return dtype, _gen


_INT_BOUNDS: dict[str, Tuple[int, int]] = {
    "tinyint": (-128, 127),
    "smallint": (-32768, 32767),
    "int": (-2_147_483_648, 2_147_483_647),
    "integer": (-2_147_483_648, 2_147_483_647),
    "bigint": (-9_223_372_036_854_775_808, 9_223_372_036_854_775_807),
    "long": (-9_223_372_036_854_775_808, 9_223_372_036_854_775_807),
}


_GENERATORS: dict[str, Generator] = {
    "string": _string_generator,
    "binary": _binary_generator,
    "boolean": _boolean_generator,
    "bool": _boolean_generator,
    "date": _date_generator,
    "timestamp": _timestamp_generator,
    "double": _floating_generator(),
    "float": _floating_generator(4),
}


def _normalise_type_hint(raw: str) -> str:
    return raw.strip().lower()


def _field_spec(field: SchemaProperty) -> tuple[T.DataType, Generator]:
    raw_type = str(field.physicalType or field.logicalType or "string")
    normalized = _normalise_type_hint(raw_type)

    if normalized.startswith("decimal"):
        return _decimal_spec(normalized)

    base = normalized.split("(", 1)[0]
    base = base.split("<", 1)[0]
    canonical = spark_type_name(base)
    dtype = _SPARK_TYPES.get(canonical)
    if dtype is None:
        return T.StringType(), _string_generator

    if canonical in _INT_BOUNDS:
        return dtype, _integer_generator(_INT_BOUNDS[canonical])

    generator = _GENERATORS.get(canonical)
    if generator:
        return dtype, generator

    return dtype, _string_generator


def generate_contract_dataset(
    spark: SparkSession,
    contract: OpenDataContractStandard,
    *,
    rows: int = 100,
    faker_locale: str | None = None,
    seed: int | None = None,
    validate_schema: bool = True,
) -> DataFrame:
    """Return a Spark ``DataFrame`` aligned to ``contract``."""

    if rows <= 0:
        raise ValueError("rows must be a positive integer")

    fake = Faker(faker_locale)
    if seed is not None:
        fake.seed_instance(seed)

    fields: List[SchemaProperty] = [prop for prop in _list_properties(contract) if prop.name]
    if not fields:
        raise ValueError("Contract does not expose any schema properties")

    struct_fields: List[T.StructField] = []
    generators: List[Generator] = []
    for field in fields:
        dtype, generator = _field_spec(field)
        struct_fields.append(T.StructField(field.name, dtype, nullable=not bool(field.required)))
        generators.append(generator)

    schema = T.StructType(struct_fields)
    data: List[tuple[object, ...]] = []
    for _ in range(rows):
        row = tuple(generator(fake, field) for generator, field in zip(generators, fields))
        data.append(row)

    df = spark.createDataFrame(data, schema=schema)

    if validate_schema:
        apply_contract(df, contract)

    return df


__all__ = ["generate_contract_dataset"]
